import numpy as np


def denorm_bbox(frame, boxes, to_int=True):
    """
    boxes: (N, 4)
        bounding box normalized coordinate (x1, y1, x2, y2)
    """
    h, w, c = frame.shape
    boxes = np.array(boxes).clip(0, 1)
    boxes = boxes * [w, h, w, h]
    if to_int:
        boxes = boxes.astype(int)
    return boxes


def denorm_point(frame, points, to_int=True):
    """
    points: (N, 2)
        points normalized coordinate (x, y)
    """
    h, w, c = frame.shape
    points = np.array(points).clip(0, 1)
    points = points * [w, h]
    if to_int:
        points = points.astype(int)
    return points


def denorm_landmark(frame, box, points, to_int=True):
    """
    box: (1, 4) or (4,)
        bounding box normalized coordinate (x1, y1, x2, y2) that contains points

    points: (N, 2)
        points coordinate normalized towards the bounding box
    """
    x1, y1, x2, y2 = box
    points = points * [x2 - x1, y2 - y1] + [x1, y1]
    points = denorm_point(frame, points, to_int)
    return points


def denorm_pixel(frame):
    if frame.dtype in [np.float32, np.float64]:
        return (frame * 255).astype(np.uint8)
    else:
        return frame



import matplotlib.pyplot as plt


def gray_imshow(image, title=None, ax=None, border=True):
    if ax is None:
        ax = plt.gca()

    ax.imshow(image, cmap="gray")
    ax = _set_border(ax, border)
    ax = _set_title(ax, title)
    return ax


def rgb_imshow(image, title=None, ax=None, border=True):
    if ax is None:
        ax = plt.gca()

    ax.imshow(image)
    ax = _set_border(ax, border)
    ax = _set_title(ax, title)
    return ax


def bgr_imshow(image, title=None, ax=None, border=True):
    return rgb_imshow(image[..., ::-1], title, ax, border)


def _set_border(ax, border):
    if border:
        ax.set(xticks=[], yticks=[])
    else:
        ax.axis("off")
    return ax


def _set_title(ax, title):
    if title is not None:
        ax.set_title(title)
    return ax





from jcopvision.utils import normalize_pixel


def enhance_brightness(image, gain=10):
    image = normalize_pixel(image)
    gain /= 255
    new_image = image + gain
    return new_image.clip(0, 1)


def enhance_contrast(image, gain=10):
    image = normalize_pixel(image)
    gain = 1 + gain / 100
    new_image = image * gain
    return new_image.clip(0, 1)



import numpy as np


def filter_bbox_by_roi(boxes, pt1=(0, 0), pt2=(1, 1)):
    boxes = np.array(boxes)
    mask = (boxes > pt1 * 2) & (boxes < pt2 * 2)
    mask = mask.all(1)
    return boxes[mask]


def filter_bbox_by_area(boxes, area_range):
    boxes = np.array(boxes)

    pt1 = boxes[:, :2]
    pt2 = boxes[:, 2:]

    area = (pt2 - pt1).prod(1)
    mask = (area > area_range[0]) & (area < area_range[1])
    return boxes[mask]


def filter_bbox_by_score(boxes, scores, min_score=0.5):
    boxes = np.array(boxes)
    mask = scores > min_score
    return boxes[mask]



import cv2
import numpy as np
from jcopvision.draw._cm import idx2color
from jcopvision.utils import denorm_point
from jcopvision.utils._denorm import denorm_pixel


def draw_circle(frame, pt, radius, color, thickness=1, is_normalized=True):
    '''
    Draw a circle


    === Input ===
    frame: array
        image / frame to be drawn

    pt: (int, int) or (float, float)
        location to draw the circle.
        If is_normalized is True, then the point will be rescaled back based on the frame size.

    color: (int, int, int)
        The circle color in BGR format

    thickness: int
        The bounding box and text thickness

    is_normalized: bool
        True if using normalized coordinate


    === Return ===
    frame: array
        annotated image / frame
    '''
    frame = frame.copy()
    frame = denorm_pixel(frame)

    if isinstance(color, int):
        color = idx2color(color)

    if is_normalized:
        pt = tuple(denorm_point(frame, pt))

    cv2.circle(frame, pt, radius, color, thickness)
    return frame



from ._nms import nms
from ._filter import filter_bbox_by_area, filter_bbox_by_roi, filter_bbox_by_score
from ._denorm import denorm_point, denorm_bbox, denorm_landmark
from ._alpha_bg import alpha_background
from ._norm import normalize_pixel



import cv2
import numpy as np


def alpha_background(w, h):
    bg = np.kron([[0.95, 0.7] * (w // 20), [0.7, 0.95] * (w // 20)] * (h // 20), np.ones((3, 10, 10)))
    bg = bg.transpose(1, 2, 0)
    return cv2.resize(bg, (w, h))



import cv2
import numpy as np
from jcopvision.draw._cm import idx2color
from jcopvision.utils import denorm_point
from jcopvision.utils._denorm import denorm_pixel


def draw_bbox(frame, pt1, pt2, color, label=None, conf=None, thickness=1, font=cv2.FONT_HERSHEY_SIMPLEX, fontsize=0.5,
              fontcolor=(0, 0, 0), is_normalized=True):
    '''
    Draw a bounding box with its label

    === Example Usage ===
    - simple usage
    draw_bbox(frame, (x1, y1), (x2, y2), color=0)

    - When you have bounding boxes with labels and confidences
    for color, ((x1, y1, x2, y2), label, conf) in enumerate(zip(boxes, labels, confs)):
        draw_bbox(frame, (x1, y1), (x2, y2), color, label, conf)


    === Input ===
    frame: array
        image / frame to be drawn

    pt1: (int, int) or (float, float)
        top left point of the bounding box.
        If is_normalized is True, then the point will be rescaled back based on the frame size.

    pt2: (int, int) or (float, float)
        bottom right point of the bounding box
        If is_normalized is True, then the point will be rescaled back based on the frame size.

    color: (int, int, int)
        The bounding box color in BGR format

    label: str
        A text or class label. It will be added to the inner top left of the bounding box

    conf: float
        The prediction confidence (0 to 1)

    thickness: int
        The bounding box and text thickness

    font: opencv's font
        The text font. Check for the available font in opencv

    fontsize: float
        The font scaling factor towards the font's base size

    fontcolor: (int, int, int)
        The text font color in BGR format

    is_normalized: bool
        True if using normalized coordinate


    === Return ===
    frame: array
        annotated image / frame
    '''
    frame = frame.copy()
    frame = denorm_pixel(frame)

    if isinstance(color, int):
        color = idx2color(color)

    if is_normalized:
        pt1 = tuple(denorm_point(frame, pt1))
        pt2 = tuple(denorm_point(frame, pt2))

    cv2.rectangle(frame, pt1, pt2, color, thickness)

    text = ""
    if label is not None:
        text += label

    if conf is not None:
        text += f" [{conf*100:.1f}%]"

    if text != "":
        # Handle text size
        (w_text, h_text), baseline = cv2.getTextSize(text, font, fontsize, thickness)

        # Filled textbox
        pt1_box = pt1
        pt2_box = (pt1[0] + w_text + 2 * baseline, pt1[1] + h_text + 2 * baseline)
        cv2.rectangle(frame, pt1_box, pt2_box, color, cv2.FILLED)

        # Add text
        pt_text = (pt1[0] + baseline, pt1[1] + h_text + baseline - thickness)
        cv2.putText(frame, text, pt_text, font, fontsize, fontcolor, thickness)
    return frame



from enum import Enum

import cv2
from jcopvision.exception import IncorrectExtensionError
from jcopvision.utils._denorm import denorm_pixel


class Interpolation(Enum):
    nearest, bilinear, bicubic, area = range(4)


class BaseWriter:
    def __init__(self, output_path, width, height, fps, fourcc, interpolation):
        self.width = int(width)
        self.height = int(height)
        self._writer = cv2.VideoWriter(output_path, fourcc, fps, (self.width, self.height))

        if interpolation in Interpolation.__members__:
            self.interpolation = Interpolation[interpolation]
        else:
            raise ValueError(f"Only supports ({', '.join(Interpolation.__members__)}) interpolation")

    def write(self, frame, mode="bgr"):
        frame = denorm_pixel(frame)

        if mode == "rgb":
            frame = frame[..., ::-1]

        h, w, c = frame.shape
        if (w != self.width) | (h != self.height):
            frame = cv2.resize(frame, (self.width, self.height))
        self._writer.write(frame)

    def close(self):
        self._writer.release()


class MP4Writer(BaseWriter):
    """
    A video writer with mp4 compression.

    === Example Usage ===
    media = MediaReader("example.mp4")
    writer = MP4Writer("output.mp4", media.width, media.height, media.frame_rate)

    === Input ===
    output_path: str
        path ended with .mp4 in its name

    width: int or float
        output video width

    height: int or float
        output video height

    fps: float
        output video frame rate

    interpolation: {nearest, bilinear, bicubic, area}
        interpolation when resizing. Only used when the input frame is different with the output shape
    """
    def __init__(self, output_path, width, height, fps, interpolation="area"):
        if not output_path.endswith(".mp4"):
            raise IncorrectExtensionError("output_path should have .mp4 extension")

        fourcc = cv2.VideoWriter_fourcc(*"MP4V")
        super().__init__(output_path, width, height, fps, fourcc, interpolation)


class AVIWriter(BaseWriter):
    """
    A video writer with avi compression.

    === Example Usage ===
    media = MediaReader("example.avi")
    writer = MP4Writer("output.avi", media.width, media.height, media.frame_rate)

    === Input ===
    output_path: str
        path ended with .avi in its name

    width: int or float
        output video width

    height: int or float
        output video height

    fps: float
        output video frame rate

    interpolation: {nearest, bilinear, bicubic, area}
        interpolation when resizing. Only used when the input frame is different with the output shape
    """
    def __init__(self, output_path, width, height, fps, interpolation="area"):
        if not output_path.endswith(".avi"):
            raise IncorrectExtensionError("output_path should have .avi extension")

        fourcc = cv2.VideoWriter_fourcc(*"MJPG")
        super().__init__(output_path, width, height, fps, fourcc, interpolation)




import cv2
import numpy as np

from jcopvision.utils import normalize_pixel


def alpha_blending(source, source_mask, background, offset=(0, 0), resize_source=None, resize_bg=None):
    # Handle tipe data
    source = normalize_pixel(source)
    source_mask = normalize_pixel(source_mask)
    background = normalize_pixel(background)

    # Handle resizing
    if resize_source is not None:
        source = cv2.resize(source.copy(), resize_source)
        source_mask = cv2.resize(source_mask.copy(), resize_source)

    if resize_bg is not None:
        background = cv2.resize(background.copy(), resize_bg)

    # Handle dimensi mask
    if source_mask.ndim == 2:
        source_mask = source_mask[:, :, None]

    # Handle offset dan inisialisasi hasil composite
    hs, ws, _ = source_mask.shape
    hb, wb, _ = background.shape
    xs = ys = xb = yb = 0

    xoff, yoff = offset
    if xoff > 0:
        xs += xoff
    else:
        xb -= xoff

    if yoff > 0:
        ys += yoff
    else:
        yb -= yoff

    w = max(xs + ws, xb + wb) - min(xs, xb)
    h = max(ys + hs, yb + hb) - min(ys, yb)
    result = np.zeros((h, w, 3))

    # Compositing
    result[yb:yb+hb, xb:xb+wb, :] = background
    result[ys:ys+hs, xs:xs+ws, :] = source_mask * source + (1 - source_mask) * result[ys:ys+hs, xs:xs+ws, :]
    return result



import PIL.Image as PillowImage
from io import BytesIO

from ipywidgets import Image
from IPython.display import display, clear_output


class JupyterCameraStream:
    def __init__(self):
        self.image = Image(width=800)

    def __call__(self, frame, mode="bgr"):
        if mode == "bgr":
            frame = frame[..., ::-1]

        clear_output(wait=True)

        f = BytesIO()
        PillowImage.fromarray(frame).save(f, "jpeg")

        self.image.value = f.getvalue()
        display(self.image)

    def set_width(self, width):
        self.image = Image(width=width)
        print(f"width is updated to {width}")


jupyter_imshow = JupyterCameraStream()



import cv2
import numpy as np

from jcopvision.io import create_sized_window, MediaReader, key_pressed
from jcopvision.utils import alpha_background, normalize_pixel


def alpha_blending(source, source_mask, background, offset=(0, 0), resize_source=None, resize_bg=None):
    # Handle tipe data
    source = normalize_pixel(source)
    source_mask = normalize_pixel(source_mask)
    background = normalize_pixel(background)

    # Handle resizing
    if resize_source is not None:
        source = cv2.resize(source.copy(), resize_source)
        source_mask = cv2.resize(source_mask.copy(), resize_source)

    if resize_bg is not None:
        background = cv2.resize(background.copy(), resize_bg)

    # Handle dimensi mask
    if source_mask.ndim == 2:
        source_mask = source_mask[:, :, None]

    # Handle offset dan inisialisasi hasil composite
    hs, ws, _ = source_mask.shape
    hb, wb, _ = background.shape
    xs = ys = xb = yb = 0

    xoff, yoff = offset
    if xoff > 0:
        xs += xoff
    else:
        xb -= xoff

    if yoff > 0:
        ys += yoff
    else:
        yb -= yoff

    w = max(xs + ws, xb + wb) - min(xs, xb)
    h = max(ys + hs, yb + hb) - min(ys, yb)
    result = np.zeros((h, w, 3))

    # Compositing
    result[yb:yb+hb, xb:xb+wb, :] = background
    result[ys:ys+hs, xs:xs+ws, :] = source_mask * source + (1 - source_mask) * result[ys:ys+hs, xs:xs+ws, :]
    return result


class ColorMaskSampler:
    def __init__(self, image_path, margin=10, app_height=500, window_name="Press `q` to quit | Press `r` to reset"):
        self.media = MediaReader(image_path)
        self.window_name = window_name
        self.app_height = app_height
        self.margin = margin

        self.mask = None
        self.image = None
        self.clicked_pixels = None
        self.result = None
        self.reset()

    def run(self):
        create_sized_window(self.app_height, self.media.aspect_ratio, self.window_name)
        cv2.setMouseCallback(self.window_name, self._sample)

        while True:
            if key_pressed("r"):
                self.reset()
            elif key_pressed("q"):
                break
            else:
                masked_image = alpha_blending(self.alpha_bg, self.mask, self.image)
                cv2.imshow(self.window_name, masked_image)

        cv2.destroyAllWindows()
        return self.result

    def reset(self):
        self.image = self.media.read()
        self.clicked_pixels = []
        self.result = ()

        h, w, c = self.image.shape
        self.mask = np.zeros((h, w, 1), dtype=np.uint8)
        self.alpha_bg = alpha_background(w, h)

    def _sample(self, action, x, y, flags, userdata):
        if action == cv2.EVENT_LBUTTONUP:
            self.clicked_pixels.append(self.image[y, x, :].astype(int))
            pmin = np.min(self.clicked_pixels, 0) - self.margin
            pmax = np.max(self.clicked_pixels, 0) + self.margin
            self.result = (pmin.clip(0, 255), pmax.clip(0, 255))
            self.mask = cv2.inRange(self.image, *self.result)



